#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import *
import os
import sys

import pytest

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from genestack_client import (FilesUtil, get_connection, make_connection_parser, get_user,
                              ExpressionNavigatorforGenes, ExpressionNavigatorforIsoforms,
                              ExpressionNavigatorforMicroarrays, AffymetrixMicroarraysNormalizationApplication,
                              SpecialFolders, GenomeQuery)

# Tests must be run on internal-dev
RNA_SEQ_GROUPS = [['GSF1431884', 'GSF1431882', 'GSF1431885'], ['GSF1431881', 'GSF1431887'],
                  ['GSF1431886', 'GSF1431883']]
ISOFORM_GROUPS = [['GSF1431850', 'GSF1431861', 'GSF1431860'], ['GSF1431852', 'GSF1431854']]
MICROARRAY_GROUPS = [['GSF10777989', 'GSF10776354'], ['GSF10776502', 'GSF10777824'], ['GSF10776014', 'GSF10774867'],
                     ['GSF10777773', 'GSF10776962']]
RAT_AFFY_ANNOTATION = "GSF14640591"
EN_TUTORIAL_FILE = 'GSF1433129'

@pytest.fixture(scope='module')
def args():
    p = make_connection_parser()
    p.add_argument('--keep', action='store_true', help='Keep created test files for manual inspection or initialization')
    return p.parse_args()


@pytest.fixture(scope='module')
def keep_files(args):
    return args.keep


@pytest.fixture(scope='module')
def conn(args):
    if get_user(args).host != "internal-dev.genestack.com":
        sys.stderr.write("Tests must be run on internal-dev")
        sys.exit(1)
    connection = get_connection(args)
    return connection


def test_en_rna_seq(conn, keep_files):
    fu = FilesUtil(conn)
    en = ExpressionNavigatorforGenes(conn)
    en_file = None
    try:
        groups = [{'accessions': accs} for accs in RNA_SEQ_GROUPS]
        en_file = en.create_file(groups, r_package=en.PKG_DESEQ, organism="new organism")
    finally:
        if (not keep_files) and (en_file is not None):
            fu.unlink_file(en_file, fu.get_special_folder(SpecialFolders.CREATED))


def test_en_isoforms(conn, keep_files):
    fu = FilesUtil(conn)
    en = ExpressionNavigatorforIsoforms(conn)
    en_file = None
    try:
        groups = [{'accessions': accs} for accs in ISOFORM_GROUPS]
        en_file = en.create_file(groups, multi_mapping_corr=True)
    finally:
        if (not keep_files) and (en_file is not None):
            fu.unlink_file(en_file, fu.get_special_folder(SpecialFolders.CREATED))


def test_en_microarrays(conn, keep_files):
    fu = FilesUtil(conn)
    en = ExpressionNavigatorforMicroarrays(conn)
    norm_app = AffymetrixMicroarraysNormalizationApplication(conn)
    en_file = None
    norm_file = None
    try:
        groups = [{'accessions': accs} for accs in MICROARRAY_GROUPS]
        groups[0]['is_control'] = True
        norm_file = norm_app.create_file([f for group in MICROARRAY_GROUPS for f in group])
        en_file = en.create_file(groups, norm_file, RAT_AFFY_ANNOTATION)
    finally:
        if not keep_files:
            created = fu.get_special_folder(SpecialFolders.CREATED)
            for f in (norm_file, en_file):
                if f is not None:
                    fu.unlink_file(f, created)


def test_get_en_stats(conn):
    en = ExpressionNavigatorforGenes(conn)
    query = GenomeQuery().set_order_ascending(True).set_limit(15).set_offset(0)
    query.set_sorting_order(GenomeQuery.SortingOrder.BY_FDR).set_contrasts(["r1", "r2"])
    query.add_filter(GenomeQuery.Filter.MAX_FDR, 0.2)
    query.add_filter(GenomeQuery.Filter.MIN_LOG_FOLD_CHANGE, 0.1)
    query.add_filter(GenomeQuery.Filter.REGULATION, GenomeQuery.Regulation.UP)

    result = en.get_differential_expression_stats({EN_TUTORIAL_FILE: query})
    entries = next(iter(result.values()))
    assert len(entries) == 30  # limit of 15 * 2 contrasts in file
    assert entries[0]['genomeFeature']['featureName'] == 'ENSG00000175745'

if __name__ == '__main__':
    pytest.main(['-v', '--tb', 'long', __file__])



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import *
from builtins import object


class Permissions(object):

    _JAVA_PKG = "com.genestack.file."

    FILE_ACCESS = _JAVA_PKG + "access"
    FILE_READ_CONTENT = _JAVA_PKG + "readContent"
    FILE_WRITE = _JAVA_PKG + "write"
    FILE_CLONE_DATA = _JAVA_PKG + "cloneData"



#!python
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import *
import os
import sys
from argparse import RawTextHelpFormatter
from datetime import datetime
from itertools import groupby
from operator import itemgetter

from genestack_client import (DataImporter, FilesUtil, GenestackServerException,
                              GenestackVersionException, SpecialFolders, get_connection,
                              make_connection_parser)

EXIT_CODE_DUPLICATED_NAMES = 23


DESCRIPTION = '''Upload raw files to server and try to auto recognize them as genestack files.

- Collecting files:
  Application can handle files and folder (will recursively collect all files).
  All paths must be valid. There is not limit to number of files.

- Uploading:
  Files are stored in subfolder of 'Raw uploads'; subfolder name corresponds
  to user local time. Files are uploaded one by one, each in multiple threads.
  In case of network errors application attempts to retry until number of retries
  exceeded (5 by default), in which case application exits with error code.
  Uploaded data is not lost though and you can continue uploading this file
  from the point you stop.

  ATTENTION: When you upload multiple files from the command line,
  be sure to remove successfully uploaded files from the arguments when before re-running
  uploader, because otherwise all of them will be uploaded to the server again.

- Recognition:
  Recognition done only if all files were uploaded successfully. It works over all files.
  Files that were not recognized are linked to subfolder 'Unrecognized files'.

  ATTENTION: Recognition of big number of files may cause server timeouts.
  Split uploading with recognition into relatively small iterations to prevent timeout
  failures.
'''

# TODO treat paths as groups for recognition:
#    each folder is separate group
#    all files joined in one group

parser = make_connection_parser()
parser.description = DESCRIPTION
parser.formatter_class = RawTextHelpFormatter
group = parser.add_argument_group("command arguments")
group.add_argument('paths',
                   help='path to files or folders',
                   metavar='<paths>', nargs='+')
group.add_argument('-n', '--no-recognition', help="don't try to recognize files", action='store_true')

exclusive_group = group.add_mutually_exclusive_group()
exclusive_group.add_argument('-F', '--folder_name', metavar='<name>',
                             help='name of the upload folder, if name is not specified it will be generated')
exclusive_group.add_argument('--upload-to', metavar='<accession>',
                             help='accession of the upload folder')


def friendly_number(number):
    """
    Produce a human-readable value for file size.
    :param number: bytes.
    :return: human-readable string.
    """
    template = '%.1f%sB'
    powers = ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    base = 1000
    number = float(number)
    for power in powers[:-1]:
        if number < base:
            return template % (number, power)
        number /= base
    return template % (number, powers[-1])


def check_duplicated_file_names(files_list):
    names = ((os.path.basename(x), x) for x in files_list)
    duplication = {}
    for file_name, collected_paths in names:
        duplication.setdefault(file_name, []).append(collected_paths)
    duplication = {k: sorted(v) for k, v in duplication.items() if len(v) > 1}
    if duplication:
        print("Files with duplicated file names were found, "
              "please rename them or load them separately to avoid confusion", file=sys.stderr)
        for name, duplicated_paths in sorted(duplication.items()):
            print(" - %s: %s" % (name, ', '.join(duplicated_paths)), file=sys.stderr)
        exit(EXIT_CODE_DUPLICATED_NAMES)


def get_files(paths):
    """
    Get file list by paths. Throws an exception if the file does not exist.

    :param paths: list of paths, can be files or directories.
    :return:
    """
    errors = []
    for path in paths:
        if not os.path.exists(path):
            errors.append(path)
    if errors:
        sys.stderr.write("Error: Path%s was not found: %s\n" % ('s' if len(paths) > 1 else '', ', '.join(paths)))
        exit(1)

    files_list = []
    total_size = 0
    for path in paths:
        if os.path.isfile(path):
            files_list.append(path)
            total_size += os.path.getsize(path)
            continue
        for base, folders, files in os.walk(path, followlinks=False):
            for f in files:
                file_path = os.path.join(base, f)
                files_list.append(file_path)
                total_size += os.path.getsize(file_path)
            for f in folders:
                folder_path = os.path.join(base, f)
                if os.path.islink(folder_path):
                    sys.stderr.write("WARNING: Symlink %s was skipped!\n" % folder_path)
    check_duplicated_file_names(files_list)
    return files_list, total_size


def upload_files(connection, files, folder_name, folder_accession):
    """
    :param genestack_client.Connection connection:
    :param list[str] files:
    :param str folder_name:
    :param str folder_accession:
    """
    importer = DataImporter(connection)
    fu = FilesUtil(connection)
    upload = fu.get_special_folder(SpecialFolders.UPLOADED)
    if not folder_accession:
        folder_name = folder_name or datetime.now().strftime('Upload %d.%m.%y %H:%M:%S')
        folder_accession = fu.create_folder(
            folder_name, parent=upload, description='Files uploaded by genestack-uploader')
    else:
        folder_name = fu.get_infos([folder_accession])[0]['name']
    accession_file_map = {}
    for f in files:
        accession = importer.load_raw(f)
        fu.link_file(accession, folder_accession)
        fu.unlink_file(accession, upload)
        accession_file_map[accession] = f
    return folder_accession, folder_name, accession_file_map


def recognize_files(connection, accession_file_map, new_folder):
    # Files Recognition
    fu = FilesUtil(connection)

    application = connection.application('genestack/upload')
    recognised_files = application.invoke('recognizeGroupsByAccession', accession_file_map.keys())

    recognized_accessions = set()
    for x in recognised_files:
        for sources in x['sourceFileInfos'].values():
            for info in sources:
                recognized_accessions.add(info['accession'])

    created_files = application.invoke('createFiles', recognised_files, [], None)
    groups = sorted(created_files['files'], key=itemgetter('kind'))
    for name, group in groupby(groups, key=itemgetter('kind')):
        print(name)
        # maybe sort by filename before printing a group?
        for f in group:
            print('\t%s / %s' % (f['accession'], f['name']))

    unrecognized_file_infos = set(accession_file_map) - recognized_accessions

    if unrecognized_file_infos:
        print('Unrecognized Raw Files')
        for accession in unrecognized_file_infos:
            print('\t%s / %s' % (accession, accession_file_map[accession].decode('utf-8')))
        # move unrecognized files to new folder
        unrecognized_folder = fu.create_folder("Unrecognized files", parent=new_folder)
        for accession in unrecognized_file_infos:
            fu.link_file(accession, unrecognized_folder)
            fu.unlink_file(accession, new_folder)
        print('Unrecognized files moved to %s / %s' % (unrecognized_folder, 'Unrecognized files'))


def main():
    args = parser.parse_args()
    files, size = get_files(args.paths)
    print('Collected %s files with total size: %s' % (len(files), friendly_number(size)))
    try:
        connection = get_connection(args)
    except GenestackVersionException as e:
        sys.stderr.write(str(e))
        sys.stderr.write('\n')
        exit(13)
    new_folder, folder_name, accessions = upload_files(connection, files, args.folder_name, args.upload_to)
    print('%s files were uploaded to %s / %s' % (len(accessions), new_folder, folder_name))
    if args.no_recognition:
        exit(0)
    try:
        recognize_files(connection, accessions, new_folder)
    except GenestackServerException as e:
        sys.stderr.write("Recognition failed: %s\n" % e)
        exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from future import standard_library

standard_library.install_aliases()
from builtins import input
from builtins import *
from builtins import object
from getpass import getpass
from urllib.parse import urlsplit

import requests

from genestack_client import Connection, GenestackAuthenticationException, GenestackException
from genestack_client.utils import isatty, interactive_select


def _get_server_url(host):
    split_host = urlsplit(host)
    has_scheme = bool(split_host.scheme)

    # compatibility with dev settings
    # TODO Add code that will do migration in configs and remove this check.
    if host.startswith('localhost'):
        return 'http://%s/frontend/endpoint' % host

    # default to HTTPS if not explicitly defined
    url_stub = host if has_scheme else 'https://%s' % host
    # trust provided path if it ends with '/frontend'
    url_stub = url_stub.rstrip('/')
    if url_stub.split('/')[-1] == 'frontend':
        return '/'.join([url_stub, 'endpoint'])
    # check both '/frontend/endpoint' and '/endpoint' and return the one that
    # works
    for path in ('frontend/', ''):
        url = '{}/{}{}'.format(url_stub, path, 'endpoint')
        # all the code later on expects base URL without trailing '/',
        # but Tomcat responds 404 for such URL, so we check with '/' appended
        if requests.get(url + '/').ok:
            return url
    raise GenestackAuthenticationException(
        "Could not connect to host '{}', check if it is defined correctly"
        "".format(host))


class User(object):
    """
    Class encapsulating all user info required for authentication.

    That includes:
     - user alias
     - server URL (or is it hostname?)
     - token *or* email/password pair
    """
    def __init__(self, email, alias=None, host=None, password=None, token=None):
        """
        All fields are optional.
        If ``alias`` is None it will be the same as ``email``.

        If you login interactively, no ``email`` or ``password`` is required.
        The alias is used to find the matching user in :py:func:`~genestack_client.get_user`

        :param email: email
        :type email: str
        :param alias: alias
        :type alias: str
        :param host: host
        :type host: str
        :param password: password
        :type password: str
        """
        self.host = host
        self.email = email
        self.password = password  # TODO make property
        self.alias = alias or email
        self.token = token

    def get_connection(self, interactive=True, debug=False, show_logs=False):
        """
        Return a logged-in connection for current user.
        If ``interactive`` is ``True`` and the password or email are unknown,
        they will be asked in interactive mode.

        :param interactive: ask email and/or password interactively.
        :type interactive: bool
        :param debug: print stack trace in case of exception
        :type debug: bool
        :param show_logs: print application logs (received from server)
        :type show_logs: bool
        :return: logged connection
        :rtype: genestack_client.Connection
        """
        connection = Connection(_get_server_url(self.host), debug=debug, show_logs=show_logs)
        if self.token:
            connection.login_by_token(self.token)
        elif self.email and self.password:
            connection.login(self.email, self.password)
        elif interactive:
            self.__interactive_login(connection)
        #else:
        #    raise GenestackException('Not enough user data to login')
        return connection

    def __repr__(self):
        return "User('%s', alias='%s', host='%s', password='%s', token='%s')" % (
            self.email, self.alias, self.host, self.password and '*****', self.token and '*****')

    def __interactive_login(self, connection):
        if not isatty():
            raise GenestackException("Interactive login is not possible")
        connection.check_version()

        email = self.email
        message = 'Connecting to %s' % self.host

        login_by_token = 'by token'
        login_by_email = 'by email and password'
        login_anonymously = 'anonymously'

        choice = interactive_select([login_by_token, login_by_email, login_anonymously],
                                    'How do you want to login')

        if choice == login_anonymously:
            return

        while True:
            if message:
                print(message)

            if choice == login_by_email:
                input_message = 'e-mail [%s]: ' % email if email and '@' in email else 'e-mail: '
                email = input(input_message).strip() or email

                password = getpass('password for %s: ' % email)
                try:
                    connection.login(email, password)
                    self.email = email
                    self.password = password
                    return
                except GenestackAuthenticationException:
                    message = ('Your username and password have been rejected by %s, '
                               'please try again' % self.host)
            else:
                token = getpass('token: ')
                try:
                    connection.login_by_token(token)
                    self.token = token
                    return
                except GenestackAuthenticationException:
                    message = 'Your token has been rejected by %s, please try again' % self.host

    def __eq__(self, other):
        return (
            isinstance(other, User) and
            self.alias == other.alias and
            self.password == other.password and
            self.host == other.host and
            self.email == other.email
        )

    def __ne__(self, other):
        return not self.__eq__(other)
